@@ -14,7 +14,7 @@ import asyncio
 import contextvars
 import functools
 import logging
-from typing import Generic, List, Any, Tuple, Callable, Optional, Type, TypeVar, Awaitable, Union
+from typing import Generic, List, Any, Tuple, Callable, Optional, Type, TypeVar, Awaitable, Union, Dict
 
 from . import conversion
 
@@ -40,19 +40,20 @@ class Connectable(Generic[T], metaclass=abc.ABCMeta):
                 receive: Optional[bool] = None,
                 read: Optional[bool] = None,
                 provide: Optional[bool] = None,
-                convert: bool = False) -> C:
+                convert: Union[bool, Tuple[Callable[[T], Any], Callable[[Any], T]]] = False) -> C:
         if isinstance(other, ConnectableWrapper):
             # If other object is not connectable itself but wraps one or more connectable objects (like, for example, a
             # `web.widgets.ValueButtonGroup`), let it use its special implementation of `connect()`.
-            other.connect(self, send=receive, receive=send, read=provide, provide=read, convert=convert)
+            other.connect(self, send=receive, receive=send, read=provide, provide=read,
+                          convert=((convert[1], convert[0]) if isinstance(convert, tuple) else convert))
         else:
-            self._connect_with(self, other, send, provide, convert)
-            self._connect_with(other, self, receive, read, convert)
+            self._connect_with(self, other, send, provide, convert[0] if isinstance(convert, tuple) else convert)
+            self._connect_with(other, self, receive, read, convert[1] if isinstance(convert, tuple) else convert)
         return self
 
     @staticmethod
     def _connect_with(source: "Connectable", target: "Connectable", send: Optional[bool], provide: Optional[bool],
-                      convert: bool):
+                      convert: Union[bool, Callable]):
         if isinstance(source, Subscribable) and isinstance(target, Writable) and (send or send is None):
             source.subscribe(target, convert=convert)
         elif send and not isinstance(source, Subscribable):
@@ -79,7 +80,7 @@ class ConnectableWrapper(Connectable[T], Generic[T], metaclass=abc.ABCMeta):
                 receive: Optional[bool] = None,
                 read: Optional[bool] = None,
                 provide: Optional[bool] = None,
-                convert: bool = False) -> C:
+                convert: Union[bool, Tuple[Callable[[T], Any], Callable[[Any], T]]] = False) -> C:
         pass
 
 
@@ -92,6 +93,10 @@ class Writable(Connectable[T], Generic[T], metaclass=abc.ABCMeta):
         classes should override *_write* instead of this method to keep profiting from the value type checking and
         magic context-based origin passing features.
 
+        This method awaits the complete transmission and processing of the new value by the next stateful
+        object/system, e.g. storing and re-publishing the value on a Variable, publishing the value from an MQTT
+        broker on an MQTT client connector, forwarding the value to all subscribers on an Expression.
+
         :param value: The new value
         :param origin: The origin / trace of the value update event, i.e. the list of objects/functions which have been
             publishing to/calling one another to cause this value update. It is used to avoid recursive feedback loops
@@ -115,7 +120,18 @@ class Writable(Connectable[T], Generic[T], metaclass=abc.ABCMeta):
         Abstract internal method containing the actual implementation-specific write-logic.
 
         It must be overridden by classes inheriting from :class:`Writable` to be updated with new values. The *_write*
-        implementation does not need to check the new value's type
+        implementation does not need to check the new value's type.
+
+        Please make sure that your `_write` implementation awaits the processing of the value update by the next
+        stateful object/system before returning:
+        - On a Variable and similar stateful re-publishing objects, which use `_stateful_publishing = True`, await the
+          storing and `_publish()`-ing of the new value
+        - On an Expression and similar stateless re-publishing objects, use `_publish_and_wait()` instead of
+          `_publish()` and await its return
+        - On external interface connectors (e.g. MQTT connector) await (at least) the successful processing of the value
+          by the external system
+
+        This is required to make the state inconsistency mitigation method of :class:`Subscribable` objects work.
 
         :param value: The new value to update this object with
         :param origin: The origin / trace of the value update event. Should be passed to :meth:`Subscribable._publish`
@@ -135,32 +151,108 @@ class UninitializedError(RuntimeError):
 
 
 class Subscribable(Connectable[T], Generic[T], metaclass=abc.ABCMeta):
+    _stateful_publishing: bool = False
+
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self._subscribers: List[Tuple[Writable[S], Optional[Callable[[T], S]]]] = []
-        self._triggers: List[LogicHandler] = []
+        self._triggers: List[Tuple[LogicHandler, bool]] = []
+        self._pending_updates: Dict[int, Set[asyncio.Task]] = {}
 
     async def __publish_write(self, subscriber: Writable[S], converter: Optional[Callable[[T], S]], value: T,
-                              origin: List[Any]):
+                              origin: List[Any], use_pending: bool):
         try:
             await subscriber.write(converter(value) if converter else value, origin + [self])  # type: ignore
         except Exception as e:
             logger.error("Error while writing new value %s from %s to %s:", value, self, subscriber, exc_info=e)
+        finally:
+            if use_pending:
+                self._pending_updates[id(subscriber)].discard(asyncio.current_task())
 
-    async def __publish_trigger(self, target: LogicHandler, value: T, origin: List[Any]):
+    async def __publish_trigger(self, target: LogicHandler, value: T,
+                                origin: List[Any], use_pending: bool):
         try:
             await target(value, origin + [self])
         except Exception as e:
             logger.error("Error while triggering %s from %s:", target, self, exc_info=e)
+        finally:
+            if use_pending:
+                self._pending_updates[id(target)].discard(asyncio.current_task())
 
-    async def _publish(self, value: T, origin: List[Any]):
-        await asyncio.gather(
-            *(self.__publish_write(subscriber, converter, value, origin)
-              for subscriber, converter in self._subscribers
-              if not any(subscriber is s for s in origin)),
-            *(self.__publish_trigger(target, value, origin)
-              for target in self._triggers)
-        )
+    def _publish(self, value: T, origin: List[Any]):
+        """
+        Method to publish a new value to all subscribers and trigger all registered logic handlers.
+
+        All logic handlers and :meth:`Writable.write` methods are called in parallel asyncio tasks, which are **not**
+        awaited to return. If `_stateful_publishing` is True on this object (or class), it will keep track of the
+        currently pending publishing tasks. This information is then used to mitigate state inconsistencies, caused by
+        value updates crossing over each other, by resetting the `origin` of following value updates to the respected
+        subscribers.
+
+        `_stateful_publishing` should be enabled for all objects that are *Subscribable* and *Writable* (i.e. allows
+        sending and receiving value updates) and have either internal state (like Variables) or represent an external,
+        stateful (or quasi-stateful) system (like a Connector object for a KNX group address/an MQTT topic). On the
+        other hand, objects, that are only *Subscribable*, should not use `_stateful_publishing`.
+
+        If the object only forwards value updates without storing a state, it should await the successful publishing of
+        the value by using the :meth:`_publish_and_wait` coroutine.
+
+        :param value: The new value to be published by this object. Must be an instance of this object's `type`.
+        :param origin: The origin list of the new value, **excluding** this object. See :ref:`base.event-origin` for
+            more details.`self` is appended automatically before calling the registered subscribers and logic handlers.
+        """
+        if self._stateful_publishing:
+            for subscriber, converter in self._subscribers:
+                reset_origin = bool(self._pending_updates[id(subscriber)])
+                if reset_origin:
+                    logger.info("Resetting origin from %s to %s; value=%s; origin=%s", self, subscriber, value, origin)
+                if reset_origin or not any(s is subscriber for s in origin):
+                    task = asyncio.create_task(self.__publish_write(subscriber, converter, value,
+                                                                    [] if reset_origin else origin, True))
+                    self._pending_updates[id(subscriber)].add(task)
+            for target, sync in self._triggers:
+                reset_origin = False
+                if sync:
+                    reset_origin = bool(self._pending_updates[id(target)])
+                task = asyncio.create_task(self.__publish_trigger(target, value, [] if reset_origin else origin, sync))
+                if sync:
+                    self._pending_updates[id(target)].add(task)
+
+        else:
+            for target, sync in self._triggers:
+                asyncio.create_task(self.__publish_trigger(target, value, origin, False))
+            for subscriber, converter in self._subscribers:
+                if not any(s is subscriber for s in origin):
+                    asyncio.create_task(self.__publish_write(subscriber, converter, value, origin, False))
+
+    async def _publish_and_wait(self, value: T, origin: List[Any]):
+        """
+        A coroutine to publish a new value to all subscribers and trigger all registered logic handlers and wait for
+        this publishing to finish.
+
+        All logic handlers and :meth:`Writable.write` methods are called in parallel asyncio tasks. However, this
+        method awaits the return of *all* them. This should only be used for **stateless** objects that re-publish
+        incoming value updates (like :ref:`Expressions <expressions>`, :class:`shc.misc.TwoWayPipe`, etc.). Other
+        objects should use :meth:`_publish` instead.
+
+        :param value: The new value to be published by this object. Must be an instance of this object's `type`.
+        :param origin: The origin list of the new value, **excluding** this object. See :ref:`base.event-origin` for
+            more details.`self` is appended automatically before calling the registered subscribers and logic handlers.
+        """
+        for target, sync in self._triggers:
+            if not sync:
+                asyncio.create_task(self.__publish_trigger(target, value, origin, False))
+        sync_jobs = [self.__publish_write(subscriber, converter, value, origin, False)
+                     for subscriber, converter in self._subscribers
+                     if not any(s is subscriber for s in origin)]
+        sync_jobs.extend(
+            self.__publish_trigger(target, value, origin, False)
+            for target, sync in self._triggers
+            if sync)
+        if len(sync_jobs) == 1:
+            await sync_jobs[0]
+        elif sync_jobs:
+            await asyncio.gather(*sync_jobs)
 
     def subscribe(self, subscriber: Writable[S], convert: Union[Callable[[T], S], bool] = False) -> None:
         """
@@ -189,8 +281,10 @@ class Subscribable(Connectable[T], Generic[T], metaclass=abc.ABCMeta):
             raise TypeError("Type mismatch of subscriber {} ({}) for {} ({})"
                             .format(repr(subscriber), subscriber.type.__name__, repr(self), self.type.__name__))
         self._subscribers.append((subscriber, converter))
+        if self._stateful_publishing:
+            self._pending_updates[id(subscriber)] = set()
 
-    def trigger(self, target: LogicHandler) -> LogicHandler:
+    def trigger(self, target: LogicHandler, synchronous: bool = False) -> LogicHandler:
         """
         Register a logic handler function to be triggered when this object is updated.
 
@@ -228,7 +322,9 @@ class Subscribable(Connectable[T], Generic[T], metaclass=abc.ABCMeta):
             mentioned above.
         :return: The ``target`` function (unchanged)
         """
-        self._triggers.append(target)
+        self._triggers.append((target, synchronous))
+        if synchronous and self._stateful_publishing:
+            self._pending_updates[id(target)] = set()
         return target
 
 
@@ -285,7 +381,7 @@ def handler(reset_origin=False, allow_recursion=False) -> Callable[[LogicHandler
 
     :param reset_origin: If True, the origin which is magically passed to all `write` calls, only contains the logic
         handler itself, not the previous `origin` list, which led to the handler's execution. This can be used to
-        change an object's value, which triggers this logic handler. This may cause infinite recursive feedback loops,
+        change an object's value, which triggered this logic handler. This may cause infinite recursive feedback loops,
         so use with care!
     :param allow_recursion: If True, recursive execution of the handler is not skipped. The handler must check the
         passed values and/or the `origin` list itself to prevent infinite feedback loops via `write` calls or calls to
@@ -299,7 +395,7 @@ def handler(reset_origin=False, allow_recursion=False) -> Callable[[LogicHandler
                     origin = magicOriginVar.get()
                 except LookupError as e:
                     raise ValueError("No origin attribute provided or set via execution context") from e
-            if any(wrapper is s for s in origin) and not allow_recursion:
+            if wrapper in origin and not allow_recursion:
                 logger.info("Skipping recursive execution of logic handler %s() via %s", f.__name__, origin)
                 return
             logger.info("Triggering logic handler %s() from %s", f.__name__, origin)
