@@ -2,17 +2,14 @@ import asyncio
 import datetime
 import unittest
 import unittest.mock
+from unittest.mock import Mock
 from typing import Optional, List
 
 from shc import timer, base
-from ._helper import ClockMock, async_test, ExampleSubscribable, AsyncMock
+from ._helper import ClockMock, async_test, ExampleSubscribable, AsyncMock, ExampleWritable
 
 
 class LogarithmicSleepTest(unittest.TestCase):
-    @classmethod
-    def setUpClass(cls) -> None:
-        ClockMock.enable()
-
     @async_test
     async def test_long_sleep(self) -> None:
         with ClockMock(datetime.datetime(2020, 1, 1, 15, 0, 0)) as clock:
@@ -41,9 +38,6 @@ class LogarithmicSleepTest(unittest.TestCase):
 
 
 class AbstractTimerTest(unittest.TestCase):
-    def setUp(self):
-        ClockMock.enable()
-
     class TestTimer(timer._AbstractScheduleTimer):
         def __init__(self, times: List[datetime.datetime]):
             super().__init__()
@@ -63,7 +57,7 @@ class AbstractTimerTest(unittest.TestCase):
                            datetime.datetime(2020, 8, 20, 22, 0, 0).astimezone()]
         events = []
 
-        async def store_time(*args, **kwargs):
+        def store_time(*args, **kwargs):
             events.append(clock_mock.now().astimezone())
 
         t = self.TestTimer(expected_events)
@@ -138,7 +132,7 @@ class EveryTimerTest(unittest.TestCase):
                                     clock.now().astimezone() + datetime.timedelta(minutes=5, seconds=-20))
             self.assertLessEqual(every_timer._next_execution(),
                                  clock.now().astimezone() + datetime.timedelta(minutes=5, seconds=20))
-    
+
     def test_aligned(self) -> None:
         with ClockMock(datetime.datetime(2020, 1, 1, 15, 7, 17)) as clock:
             every_timer = timer.Every(datetime.timedelta(minutes=5), align=True)
@@ -270,14 +264,10 @@ class AtTimerTest(unittest.TestCase):
 
     def test_exception(self) -> None:
         with self.assertRaises(ValueError):
-            once_timer = timer.At(day=[1,5,15], weeknum=timer.EveryNth(2))
+            once_timer = timer.At(day=[1, 5, 15], weeknum=timer.EveryNth(2))
 
 
 class BoolTimerTest(unittest.TestCase):
-    @classmethod
-    def setUpClass(cls) -> None:
-        ClockMock.enable()
-
     @async_test
     async def test_ton(self) -> None:
         begin = datetime.datetime(2020, 12, 31, 23, 59, 46)
@@ -289,7 +279,7 @@ class BoolTimerTest(unittest.TestCase):
         base = ExampleSubscribable(bool)
         ton = timer.TOn(base, datetime.timedelta(seconds=42))
 
-        with unittest.mock.patch.object(ton, "_publish", new=AsyncMock(side_effect=save_time)) as publish_mock:
+        with unittest.mock.patch.object(ton, "_publish", new=Mock(side_effect=save_time)) as publish_mock:
             with ClockMock(begin, actual_sleep=0.05) as clock:
                 self.assertFalse(await ton.read())
 
@@ -341,7 +331,7 @@ class BoolTimerTest(unittest.TestCase):
         base = ExampleSubscribable(bool)
         toff = timer.TOff(base, datetime.timedelta(seconds=42))
 
-        with unittest.mock.patch.object(toff, "_publish", new=AsyncMock(side_effect=save_time)) as publish_mock:
+        with unittest.mock.patch.object(toff, "_publish", new=Mock(side_effect=save_time)) as publish_mock:
             with ClockMock(begin, actual_sleep=0.05) as clock:
                 self.assertFalse(await toff.read())
 
@@ -395,7 +385,7 @@ class BoolTimerTest(unittest.TestCase):
         base = ExampleSubscribable(bool)
         tpulse = timer.TPulse(base, datetime.timedelta(seconds=42))
 
-        with unittest.mock.patch.object(tpulse, "_publish", new=AsyncMock(side_effect=save_time)) as publish_mock:
+        with unittest.mock.patch.object(tpulse, "_publish", new=Mock(side_effect=save_time)) as publish_mock:
             with ClockMock(begin, actual_sleep=0.05) as clock:
                 self.assertFalse(await tpulse.read())
 
@@ -414,34 +404,170 @@ class BoolTimerTest(unittest.TestCase):
                 self.assertAlmostEqual(begin + datetime.timedelta(seconds=42.01), call_times[-1],
                                        delta=datetime.timedelta(seconds=.01))
 
-                # Pulse should not be stoppable with False
+                # Pulse should not be stoppable with False or extendable with a second rising edge
                 publish_mock.reset_mock()
+                await base.publish(False, [self])
+                await asyncio.sleep(0.01)
+                publish_mock.assert_not_called()
                 start = clock.now()
                 await base.publish(True, [self])
                 await asyncio.sleep(0.01)
                 publish_mock.assert_called_with(True, unittest.mock.ANY)
+                # second pulse after 20s
                 await asyncio.sleep(20)
                 await base.publish(False, [self])
                 await asyncio.sleep(0.01)
                 self.assertEqual(1, publish_mock.call_count)
+                await asyncio.sleep(20)
+                await base.publish(True, [self])
+                await asyncio.sleep(0.01)
+                self.assertEqual(1, publish_mock.call_count)
+                # Wait 25s more
+                await asyncio.sleep(25)
+                self.assertEqual(2, publish_mock.call_count)
+                publish_mock.assert_called_with(False, unittest.mock.ANY)
+                self.assertAlmostEqual(start + datetime.timedelta(seconds=42.01), call_times[-1],
+                                       delta=datetime.timedelta(seconds=.01))
+
+                # And a two consecutive True values should not trigger a second pulse
+                publish_mock.reset_mock()
+                await base.publish(True, [self])
+                await asyncio.sleep(0.2)
+                publish_mock.assert_not_called()
+
+                # ... but a False -> True edge should
+                await base.publish(False, [self])
+                await asyncio.sleep(0.2)
+                publish_mock.assert_not_called()
+                start = clock.now()
+                await base.publish(True, [self])
+                await asyncio.sleep(0.01)
+                self.assertEqual(1, publish_mock.call_count)
                 await asyncio.sleep(45)
                 self.assertEqual(2, publish_mock.call_count)
                 publish_mock.assert_called_with(False, unittest.mock.ANY)
                 self.assertAlmostEqual(start + datetime.timedelta(seconds=42.01), call_times[-1],
                                        delta=datetime.timedelta(seconds=.01))
 
-                # ... or extendable with True
+
+class TimerSwitchTest(unittest.TestCase):
+    @async_test
+    async def test_simple(self) -> None:
+        pub_on1 = ExampleSubscribable(type(None))
+        pub_on2 = ExampleSubscribable(type(None))
+        pub_off1 = ExampleSubscribable(type(None))
+        pub_off2 = ExampleSubscribable(type(None))
+
+        timerswitch = timer.TimerSwitch([pub_on1, pub_on2], [pub_off1, pub_off2])
+
+        with unittest.mock.patch.object(timerswitch, "_publish") as publish_mock:
+            self.assertFalse(await timerswitch.read())
+
+            await pub_on1.publish(None, [self])
+            publish_mock.assert_called_once_with(True, unittest.mock.ANY)
+            self.assertTrue(await timerswitch.read())
+
+            publish_mock.reset_mock()
+            await pub_on2.publish(None, [self])
+            publish_mock.assert_not_called()
+            self.assertTrue(await timerswitch.read())
+
+            await pub_off2.publish(None, [self])
+            publish_mock.assert_called_once_with(False, unittest.mock.ANY)
+            self.assertFalse(await timerswitch.read())
+
+            publish_mock.reset_mock()
+            await pub_off1.publish(None, [self])
+            publish_mock.assert_not_called()
+            self.assertFalse(await timerswitch.read())
+
+    @async_test
+    async def test_duration(self) -> None:
+        begin = datetime.datetime(2020, 1, 1, 0, 0, 0)
+        pub_on1 = ExampleSubscribable(type(None))
+        pub_on2 = ExampleSubscribable(type(None))
+
+        timerswitch = timer.TimerSwitch([pub_on1, pub_on2], duration=datetime.timedelta(seconds=42))
+
+        with unittest.mock.patch.object(timerswitch, "_publish") as publish_mock:
+            with ClockMock(begin, actual_sleep=0.05) as clock:
+                self.assertFalse(await timerswitch.read())
+
+                await pub_on2.publish(None, [self])
+                publish_mock.assert_called_once_with(True, unittest.mock.ANY)
+                self.assertTrue(await timerswitch.read())
+
+                # Retrigger duration after 32 seconds
+                await asyncio.sleep(32)
                 publish_mock.reset_mock()
-                start = clock.now()
-                await base.publish(True, [self])
-                await asyncio.sleep(0.01)
-                publish_mock.assert_called_with(True, unittest.mock.ANY)
+                await pub_on1.publish(None, [self])
+                publish_mock.assert_not_called()
+                self.assertTrue(await timerswitch.read())
+
+                # After 52 seconds, the timerswitch should still be on, since it has been retriggered at 32 sec.
                 await asyncio.sleep(20)
-                await base.publish(True, [self])
-                await asyncio.sleep(0.01)
-                self.assertEqual(1, publish_mock.call_count)
-                await asyncio.sleep(45)
-                self.assertEqual(2, publish_mock.call_count)
-                publish_mock.assert_called_with(False, unittest.mock.ANY)
-                self.assertAlmostEqual(start + datetime.timedelta(seconds=42.01), call_times[-1],
-                                       delta=datetime.timedelta(seconds=.01))
+                publish_mock.assert_not_called()
+                self.assertTrue(await timerswitch.read())
+
+                # After 74 seconds the timerswitch should automatically switch off. Let't test its state at 73 and 75
+                # sec.
+                await asyncio.sleep(21)
+                publish_mock.assert_not_called()
+                self.assertTrue(await timerswitch.read())
+
+                await asyncio.sleep(2)
+                publish_mock.assert_called_once_with(False, unittest.mock.ANY)
+                self.assertFalse(await timerswitch.read())
+
+    def test_error(self) -> None:
+        pub_on1 = ExampleSubscribable(bool)
+        pub_off1 = ExampleSubscribable(bool)
+
+        with self.assertRaises(ValueError):
+            timer.TimerSwitch([pub_on1], [pub_off1], datetime.timedelta(seconds=42))
+
+
+class RateLimitedSubscriptionTest(unittest.TestCase):
+    @async_test
+    async def test_simple(self) -> None:
+        begin = datetime.datetime(2020, 12, 31, 23, 59, 46)
+        call_times = []
+
+        def save_time(*args):
+            call_times.append(datetime.datetime.now())
+
+        base = ExampleSubscribable(int)
+        rate_limiter = timer.RateLimitedSubscription(base, 1.0)
+
+        self.assertIs(rate_limiter.type, int)
+
+        with unittest.mock.patch.object(rate_limiter, "_publish", new=Mock(side_effect=save_time)) as publish_mock:
+            with ClockMock(begin, actual_sleep=0.05) as clock:
+                # First value should be forwarded immediately
+                await base.publish(42, [self])
+                await asyncio.sleep(0.1)
+                publish_mock.assert_called_once_with(42, unittest.mock.ANY)
+                publish_mock.reset_mock()
+                await asyncio.sleep(0.4)
+
+                # Second value within min_interval should be hold back for remaining interval time (1-0.1-0.4 = 0.5)
+                # and be superseded by third value
+                await base.publish(56, [self])
+                await asyncio.sleep(0.1)
+                publish_mock.assert_not_called()
+
+                await base.publish(21, [unittest.mock.sentinel, self])
+                await asyncio.sleep(0.5)
+                publish_mock.assert_called_once_with(21, unittest.mock.ANY)
+                self.assertEqual(21, publish_mock.call_args[0][0])
+                self.assertIs(publish_mock.call_args[0][1][0], unittest.mock.sentinel)
+
+                # After longer period of time there should not be new calls
+                await asyncio.sleep(5)
+                publish_mock.assert_called_once()
+
+                # A new value should now be forwarded immediately, again
+                publish_mock.reset_mock()
+                await base.publish(42, [self])
+                await asyncio.sleep(0.1)
+                publish_mock.assert_called_once_with(42, unittest.mock.ANY)
