@@ -18,18 +18,39 @@ from selenium.webdriver.common.alert import Alert  # type: ignore
 from selenium.webdriver.common.keys import Keys  # type: ignore
 from selenium.webdriver import ActionChains  # type: ignore
 
-from shc import web
+import shc.web
+import shc.web.widgets
 from shc.datatypes import RangeFloat1, RGBUInt8, RangeUInt8
+from shc.supervisor import AbstractInterface, InterfaceStatus, ServiceStatus
 from ._helper import InterfaceThreadRunner, ExampleReadable, AsyncMock, async_test
 
 
+class StatusTestInterface(AbstractInterface):
+    def __init__(self, name):
+        super().__init__()
+        self.status = InterfaceStatus()
+        self.name = name
+
+    async def start(self) -> None:
+        pass
+
+    async def stop(self) -> None:
+        pass
+
+    async def get_status(self) -> "InterfaceStatus":
+        return self.status
+
+    def __repr__(self):
+        return f"StatusTestInterface({self.name})"
+
+
 @unittest.skipIf(shutil.which("geckodriver") is None, "Selenium's geckodriver is not available in PATH")
 class AbstractWebTest(unittest.TestCase):
     driver: webdriver.Firefox
 
     def setUp(self) -> None:
-        self.server = web.WebServer("localhost", 42080, 'index')
-        self.server_runner = InterfaceThreadRunner(self.server)
+        self.server_runner = InterfaceThreadRunner(shc.web.WebServer, "localhost", 42080, 'index')
+        self.server = self.server_runner.interface
 
     @classmethod
     def setUpClass(cls) -> None:
@@ -52,12 +73,12 @@ class SimpleWebTest(AbstractWebTest):
 
     def test_page(self) -> None:
         page = self.server.page('index', 'Home Page')
-        page.add_item(web.widgets.ButtonGroup("My button group", [
-            web.widgets.StatelessButton(42, "Foobar")
+        page.add_item(shc.web.widgets.ButtonGroup("My button group", [
+            shc.web.widgets.StatelessButton(42, "Foobar")
         ]))
         page.new_segment("Another segment", full_width=True)
-        page.add_item(web.widgets.ButtonGroup("Another button group", [
-            web.widgets.StatelessButton(42, "Bar")
+        page.add_item(shc.web.widgets.ButtonGroup("Another button group", [
+            shc.web.widgets.StatelessButton(42, "Bar")
         ]))
 
         self.server_runner.start()
@@ -96,22 +117,77 @@ class SimpleWebTest(AbstractWebTest):
         submenu_entry.find_element_by_css_selector('i.bars.icon')
 
 
+class MonitoringTest(unittest.TestCase):
+    def setUp(self) -> None:
+        shc.supervisor._REGISTERED_INTERFACES.clear()
+        self.interface1 = StatusTestInterface("Interface 1")
+        self.interface1.criticality = shc.supervisor.ServiceCriticality.INFO
+        self.interface2 = StatusTestInterface("Interface 2")
+        self.interface2.criticality = shc.supervisor.ServiceCriticality.WARNING
+        self.interface3 = StatusTestInterface("Interface 3")
+        self.interface3.criticality = shc.supervisor.ServiceCriticality.CRITICAL
+        self.server_runner = InterfaceThreadRunner(shc.web.WebServer, "localhost", 42080, 'index')
+        self.server = self.server_runner.interface
+
+    def tearDown(self) -> None:
+        self.server_runner.stop()
+        shc.supervisor._REGISTERED_INTERFACES.clear()
+
+    @async_test
+    async def test_monitoring_json(self) -> None:
+        self.server_runner.start()
+
+        self.interface1.status = InterfaceStatus(ServiceStatus.CRITICAL, "Something is wrong", {"badness": 100})
+        self.interface3.status = InterfaceStatus(ServiceStatus.WARNING, "Be warned",
+                                                 {"info": "intervention may be required"})
+        headers = {"Accept": "application/json"}
+        async with aiohttp.ClientSession(headers=headers) as session:
+            async with session.get('http://localhost:42080/monitoring') as resp:
+                self.assertEqual(213, resp.status)
+                data = await resp.json()
+
+            self.assertEqual(1, data['status'])
+            self.assertEqual(2, data['interfaces']['StatusTestInterface(Interface 1)']['status'])
+            self.assertEqual("Something is wrong", data['interfaces']['StatusTestInterface(Interface 1)']['message'])
+            self.assertEqual(100, data['interfaces']['StatusTestInterface(Interface 1)']['indicators']['badness'])
+            self.assertEqual(0, data['interfaces']['StatusTestInterface(Interface 2)']['status'])
+            self.assertEqual("", data['interfaces']['StatusTestInterface(Interface 2)']['message'])
+            self.assertEqual(1, data['interfaces']['StatusTestInterface(Interface 3)']['status'])
+
+            self.interface3.status = InterfaceStatus(ServiceStatus.CRITICAL, "ERROR ERROR!!1!")
+
+            async with session.get('http://localhost:42080/monitoring') as resp:
+                self.assertEqual(513, resp.status)
+                data = await resp.json()
+                self.assertEqual(2, data['status'])
+
+            self.interface3.status = InterfaceStatus(ServiceStatus.OK, "Service restored")
+
+            async with session.get('http://localhost:42080/monitoring') as resp:
+                self.assertEqual(200, resp.status)
+                data = await resp.json()
+                self.assertEqual(0, data['status'])
+
+    # TODO test HTML/UI monitoring page
+
+
 class WebWidgetsTest(AbstractWebTest):
     def test_switch(self) -> None:
         page = self.server.page('index')
-        switch_widget = web.widgets.Switch("Main Power").connect(ExampleReadable(bool, True))
+        switch_widget = shc.web.widgets.Switch("Main Power").connect(ExampleReadable(bool, True))
         page.add_item(switch_widget)
 
-        with unittest.mock.patch.object(switch_widget, '_publish', new_callable=AsyncMock) as publish_mock:
+        with unittest.mock.patch.object(switch_widget, '_publish') as publish_mock:
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
-            time.sleep(0.05)
+            time.sleep(0.4)
             checkbox_element = self.driver.find_element_by_xpath(
                 '//*[normalize-space(text()) = "Main Power"]/..//input')
             self.assertTrue(checkbox_element.is_selected())
 
             asyncio.run_coroutine_threadsafe(switch_widget.write(False, [self]), loop=self.server_runner.loop).result()
             time.sleep(0.05)
+            publish_mock.reset_mock()
             self.assertFalse(checkbox_element.is_selected())
 
             checkbox_container = checkbox_element.find_element_by_xpath('./..')
@@ -121,13 +197,13 @@ class WebWidgetsTest(AbstractWebTest):
             publish_mock.assert_called_once_with(True, unittest.mock.ANY)
 
     def test_switch_confirm(self) -> None:
-        switch = web.widgets.Switch("Some Switch", confirm_values=(True,), confirm_message="My text")\
+        switch = shc.web.widgets.Switch("Some Switch", confirm_values=(True,), confirm_message="My text")\
             .connect(ExampleReadable(bool, False))
 
         page = self.server.page('index')
         page.add_item(switch)
 
-        with unittest.mock.patch.object(switch, '_publish', new_callable=AsyncMock) as publish_mock:
+        with unittest.mock.patch.object(switch, '_publish') as publish_mock:
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
@@ -151,20 +227,20 @@ class WebWidgetsTest(AbstractWebTest):
             publish_mock.assert_called_once_with(False, unittest.mock.ANY)
 
     def test_buttons(self) -> None:
-        b1 = web.widgets.ToggleButton(label="B1", color='yellow')
-        b2: web.widgets.DisplayButton[bool] = web.widgets.DisplayButton(label="B2", color='blue')
-        b3 = web.widgets.StatelessButton(42, "B3")
-        b4 = web.widgets.ValueButton(42, "B4", color="red")
+        b1 = shc.web.widgets.ToggleButton(label="B1", color='yellow')
+        b2: shc.web.widgets.DisplayButton[bool] = shc.web.widgets.DisplayButton(label="B2", color='blue')
+        b3 = shc.web.widgets.StatelessButton(42, "B3")
+        b4 = shc.web.widgets.ValueButton(42, "B4", color="red")
         ExampleReadable(bool, True).connect(b1)
         ExampleReadable(bool, True).connect(b2)
         ExampleReadable(int, 42).connect(b4)
 
         page = self.server.page('index')
-        page.add_item(web.widgets.ButtonGroup("My button group", [b1, b2, b3, b4]))
+        page.add_item(shc.web.widgets.ButtonGroup("My button group", [b1, b2, b3, b4]))
 
-        with unittest.mock.patch.object(b1, '_publish', new_callable=AsyncMock) as b1_publish,\
-                unittest.mock.patch.object(b3, '_publish', new_callable=AsyncMock) as b3_publish,\
-                unittest.mock.patch.object(b4, '_publish', new_callable=AsyncMock) as b4_publish:
+        with unittest.mock.patch.object(b1, '_publish') as b1_publish,\
+                unittest.mock.patch.object(b3, '_publish') as b3_publish,\
+                unittest.mock.patch.object(b4, '_publish') as b4_publish:
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
@@ -189,6 +265,9 @@ class WebWidgetsTest(AbstractWebTest):
             asyncio.run_coroutine_threadsafe(b2.write(False, [self]), loop=self.server_runner.loop).result()
             asyncio.run_coroutine_threadsafe(b4.write(56, [self]), loop=self.server_runner.loop).result()
             time.sleep(0.05)
+            b1_publish.reset_mock()
+            b3_publish.reset_mock()
+            b4_publish.reset_mock()
             self.assertNotIn('blue', b2_element.get_attribute('class'))
             self.assertNotIn('red', b4_element.get_attribute('class'))
 
@@ -212,15 +291,16 @@ class WebWidgetsTest(AbstractWebTest):
             b4_publish.assert_called_once_with(42, unittest.mock.ANY)
 
     def test_button_confirm(self) -> None:
-        button = web.widgets.ToggleButton(label="B1", color='yellow', confirm_message="Sure?", confirm_values=(True,))
+        button = shc.web.widgets.ToggleButton(label="B1", color='yellow', confirm_message="Sure?",
+                                              confirm_values=(True,))
 
         page = self.server.page('index')
-        page.add_item(web.widgets.ButtonGroup("My button group", [button]))
+        page.add_item(shc.web.widgets.ButtonGroup("My button group", [button]))
 
-        with unittest.mock.patch.object(button, '_publish', new_callable=AsyncMock) as publish_mock:
+        with unittest.mock.patch.object(button, '_publish') as publish_mock:
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
-            time.sleep(0.05)
+            time.sleep(0.4)
 
             button_element = self.driver.find_element_by_xpath('//button[normalize-space(text()) = "B1"]')
 
@@ -256,7 +336,7 @@ class WebWidgetsTest(AbstractWebTest):
 
     def test_display(self) -> None:
         page = self.server.page('index')
-        text_widget = web.widgets.TextDisplay(int, "{} lux", "Brightness").connect(ExampleReadable(int, 42))
+        text_widget = shc.web.widgets.TextDisplay(int, "{} lux", "Brightness").connect(ExampleReadable(int, 42))
         page.add_item(text_widget)
 
         self.server_runner.start()
@@ -271,10 +351,10 @@ class WebWidgetsTest(AbstractWebTest):
 
     def test_input_int(self) -> None:
         page = self.server.page('index')
-        input_widget = web.widgets.TextInput(int, "Brightness").connect(ExampleReadable(int, 42))
+        input_widget = shc.web.widgets.TextInput(int, "Brightness").connect(ExampleReadable(int, 42))
         page.add_item(input_widget)
 
-        with unittest.mock.patch.object(input_widget, '_publish', new_callable=AsyncMock) as publish_mock:
+        with unittest.mock.patch.object(input_widget, '_publish') as publish_mock:
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
@@ -283,6 +363,7 @@ class WebWidgetsTest(AbstractWebTest):
 
             asyncio.run_coroutine_threadsafe(input_widget.write(56, [self]), loop=self.server_runner.loop).result()
             time.sleep(0.05)
+            publish_mock.reset_mock()
             self.assertEqual("56", input_element.get_attribute("value"))
 
             input_element.send_keys(Keys.SHIFT + Keys.HOME, Keys.BACK_SPACE)
@@ -305,10 +386,11 @@ class WebWidgetsTest(AbstractWebTest):
 
     def test_input_string(self) -> None:
         page = self.server.page('index')
-        input_widget = web.widgets.TextInput(str, "Message of the Day").connect(ExampleReadable(str, "Hello, World!"))
+        input_widget = shc.web.widgets.TextInput(str, "Message of the Day")\
+            .connect(ExampleReadable(str, "Hello, World!"))
         page.add_item(input_widget)
 
-        with unittest.mock.patch.object(input_widget, '_publish', new_callable=AsyncMock) as publish_mock:
+        with unittest.mock.patch.object(input_widget, '_publish') as publish_mock:
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
@@ -316,8 +398,10 @@ class WebWidgetsTest(AbstractWebTest):
                 '//*[normalize-space(text()) = "Message of the Day"]/..//input')
             self.assertEqual("Hello, World!", input_element.get_attribute("value"))
 
-            asyncio.run_coroutine_threadsafe(input_widget.write("Foobar", [self]), loop=self.server_runner.loop).result()
+            asyncio.run_coroutine_threadsafe(input_widget.write("Foobar", [self]), loop=self.server_runner.loop)\
+                .result()
             time.sleep(0.05)
+            publish_mock.reset_mock()
             self.assertEqual("Foobar", input_element.get_attribute("value"))
 
             input_element.send_keys(Keys.SHIFT + Keys.HOME, Keys.BACK_SPACE)
@@ -328,10 +412,10 @@ class WebWidgetsTest(AbstractWebTest):
 
     def test_slider(self) -> None:
         page = self.server.page('index')
-        input_widget = web.widgets.Slider("Amount of Foo").connect(ExampleReadable(RangeFloat1, RangeFloat1(0.3)))
+        input_widget = shc.web.widgets.Slider("Amount of Foo").connect(ExampleReadable(RangeFloat1, RangeFloat1(0.3)))
         page.add_item(input_widget)
 
-        with unittest.mock.patch.object(input_widget, '_publish', new_callable=AsyncMock) as publish_mock:
+        with unittest.mock.patch.object(input_widget, '_publish') as publish_mock:
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
@@ -362,13 +446,13 @@ class WebWidgetsTest(AbstractWebTest):
 
     def test_hiderow(self) -> None:
         page = self.server.page('index')
-        foo_button = web.widgets.StatelessButton(True, web.widgets.icon('power off'))
-        foo_row = web.widgets.HideRow("Foo", foo_button, 'red').connect(ExampleReadable(bool, True))
-        bar_row = web.widgets.HideRow("Bar").connect(ExampleReadable(bool, True))
-        foobar_row = web.widgets.HideRow("Foobar", color='yellow').connect(ExampleReadable(bool, False))
-        page.add_item(web.widgets.HideRowBox([foo_row, bar_row, foobar_row]))
+        foo_button = shc.web.widgets.StatelessButton(True, shc.web.widgets.icon('power off'))
+        foo_row = shc.web.widgets.HideRow("Foo", foo_button, 'red').connect(ExampleReadable(bool, True))
+        bar_row = shc.web.widgets.HideRow("Bar").connect(ExampleReadable(bool, True))
+        foobar_row = shc.web.widgets.HideRow("Foobar", color='yellow').connect(ExampleReadable(bool, False))
+        page.add_item(shc.web.widgets.HideRowBox([foo_row, bar_row, foobar_row]))
 
-        with unittest.mock.patch.object(foo_button, '_publish', new_callable=AsyncMock) as publish_mock:
+        with unittest.mock.patch.object(foo_button, '_publish') as publish_mock:
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
@@ -403,11 +487,11 @@ class WebWidgetsTest(AbstractWebTest):
 
     def test_colorchoser(self) -> None:
         page = self.server.page('index')
-        input_widget = web.widgets.ColorChoser()\
+        input_widget = shc.web.widgets.ColorChoser()\
             .connect(ExampleReadable(RGBUInt8, RGBUInt8(RangeUInt8(127), RangeUInt8(127), RangeUInt8(127))))
         page.add_item(input_widget)
 
-        with unittest.mock.patch.object(input_widget, '_publish', new_callable=AsyncMock) as publish_mock:
+        with unittest.mock.patch.object(input_widget, '_publish') as publish_mock:
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
             time.sleep(0.4)
@@ -460,14 +544,14 @@ class WebWidgetsTest(AbstractWebTest):
             YET_ANOTHER_VALUE = 2
 
         page = self.server.page('index')
-        input_widget = web.widgets.EnumSelect(ExampleEnum, "Select the Foo")\
+        input_widget = shc.web.widgets.EnumSelect(ExampleEnum, "Select the Foo")\
             .connect(ExampleReadable(ExampleEnum, ExampleEnum.SOME_OTHER_VALUE))
         page.add_item(input_widget)
 
-        with unittest.mock.patch.object(input_widget, '_publish', new_callable=AsyncMock) as publish_mock:
+        with unittest.mock.patch.object(input_widget, '_publish') as publish_mock:
             self.server_runner.start()
             self.driver.get("http://localhost:42080")
-            time.sleep(0.05)
+            time.sleep(0.4)
             container_element = self.driver.find_element_by_xpath(
                 '//*[normalize-space(text()) = "Select the Foo"]/..')
             menu_element = container_element.find_element_by_css_selector('.selection.dropdown')
@@ -490,10 +574,10 @@ class WebWidgetsTest(AbstractWebTest):
 
     def test_image_map(self) -> None:
         page = self.server.page('index')
-        b1 = web.widgets.ToggleButton(label="B1", color='yellow').connect(ExampleReadable(bool, True))
-        l2 = web.widgets.ImageMapLabel(float, color='red').connect(ExampleReadable(float, 15.3))
+        b1 = shc.web.widgets.ToggleButton(label="B1", color='yellow').connect(ExampleReadable(bool, True))
+        l2 = shc.web.widgets.ImageMapLabel(float, color='red').connect(ExampleReadable(float, 15.3))
 
-        page.add_item(web.widgets.ImageMap((Path(__file__)).parent / 'assets' / 'example_image.jpg', [
+        page.add_item(shc.web.widgets.ImageMap((Path(__file__)).parent / 'assets' / 'example_image.jpg', [
             (0.3, 0.3, b1),
             (0.9, 0.55, l2)
         ]))
@@ -535,8 +619,8 @@ class TestAPI(unittest.TestCase):
     # another HTTP implementation and have a more realistic control flow/timing (with different threads instead of one
     # AsyncIO event loop)
     def setUp(self) -> None:
-        self.server = web.WebServer("localhost", 42080, 'index')
-        self.server_runner = InterfaceThreadRunner(self.server)
+        self.server_runner = InterfaceThreadRunner(shc.web.WebServer, "localhost", 42080, 'index')
+        self.server = self.server_runner.interface
 
     def tearDown(self) -> None:
         self.server_runner.stop()
@@ -568,7 +652,7 @@ class TestAPI(unittest.TestCase):
 
         async def scheduled_update(value):
             await asyncio.sleep(0.2)
-            await api_object.write(value, self)
+            await api_object.write(value, [self])
 
         tic = time.time()
         asyncio.run_coroutine_threadsafe(scheduled_update(56), self.server_runner.loop)
@@ -598,7 +682,7 @@ class TestAPI(unittest.TestCase):
 
         async def scheduled_update(value):
             await asyncio.sleep(0.2)
-            await api_object.write(value, self)
+            await api_object.write(value, [self])
 
         # A GET request with matching ETag and wait parameter. It should wait for the next value
         tic = time.time()
@@ -629,7 +713,7 @@ class TestAPI(unittest.TestCase):
         self.server_runner.start()
 
         # POST to non-existing object
-        with unittest.mock.patch.object(api_object, '_publish', new_callable=AsyncMock) as publish_mock:
+        with unittest.mock.patch.object(api_object, '_publish') as publish_mock:
             with self.assertRaises(urllib.error.HTTPError) as cm:
                 request = urllib.request.Request("http://localhost:42080/api/v1/object/non_existing_object",
                                                  data=json.dumps(56).encode(), method="POST")
@@ -638,7 +722,7 @@ class TestAPI(unittest.TestCase):
         publish_mock.assert_not_called()
 
         # valid POST
-        with unittest.mock.patch.object(api_object, '_publish', new_callable=AsyncMock) as publish_mock:
+        with unittest.mock.patch.object(api_object, '_publish') as publish_mock:
             request = urllib.request.Request("http://localhost:42080/api/v1/object/the_api_object",
                                              data=json.dumps(56).encode(), method="POST")
             urllib.request.urlopen(request)
@@ -661,8 +745,8 @@ class TestAPI(unittest.TestCase):
 
 class WebSocketAPITest(unittest.TestCase):
     def setUp(self) -> None:
-        self.server = web.WebServer("localhost", 42080, 'index')
-        self.server_runner = InterfaceThreadRunner(self.server)
+        self.server_runner = InterfaceThreadRunner(shc.web.WebServer, "localhost", 42080, 'index')
+        self.server = self.server_runner.interface
 
         self.closing = False
         self.ws_callback = unittest.mock.Mock()
@@ -769,7 +853,7 @@ class WebSocketAPITest(unittest.TestCase):
         self.server_runner.start()
         await self.start_websocket()
 
-        with unittest.mock.patch.object(api_object, '_publish', new_callable=AsyncMock) as publish_mock:
+        with unittest.mock.patch.object(api_object, '_publish') as publish_mock:
             await self.ws.send_json({'action': 'post', 'name': 'the_api_object', 'value': 56})
             await asyncio.sleep(0.05)
 
@@ -783,6 +867,7 @@ class WebSocketAPITest(unittest.TestCase):
     @async_test
     async def test_subscribe(self) -> None:
         api_object = self.server.api(int, "the_api_object").connect(ExampleReadable(int, 42))
+        api_object2 = self.server.api(int, "the_other_api_object")
         self.server_runner.start()
         await self.start_websocket()
 
@@ -797,11 +882,23 @@ class WebSocketAPITest(unittest.TestCase):
         self.assertEqual(200, data['status'])
 
         self.ws_callback.reset_mock()
-        asyncio.run_coroutine_threadsafe(api_object.write(56, [self]), self.server_runner.loop)
+        await self.ws.send_json({'action': 'subscribe', 'name': 'the_other_api_object', 'handle': 42})
         await asyncio.sleep(0.05)
 
         self.ws_callback.assert_called_once()
         data = json.loads(self.ws_callback.call_args[0][0])
-        self.assertEqual('the_api_object', data['name'])
+        self.assertEqual('subscribe', data['action'])
+        self.assertEqual('the_other_api_object', data['name'])
+        self.assertEqual(42, data['handle'])
+        self.assertNotIn('value', data)
+
+        self.ws_callback.reset_mock()
+        asyncio.run_coroutine_threadsafe(api_object.write(56, [self]), self.server_runner.loop)
+        asyncio.run_coroutine_threadsafe(api_object2.write(56, [self]), self.server_runner.loop)
+        await asyncio.sleep(0.05)
+
+        self.assertEqual(2, self.ws_callback.call_count)
+        data = json.loads(self.ws_callback.call_args[0][0])
+        self.assertIn(data['name'], ('the_api_object', 'the_other_api_object'))
         self.assertEqual(56, data['value'])
         self.assertEqual(200, data['status'])
